<?php

require __DIR__ . '/dep/vendor/autoload.php';

include_once(     "CSV.php" );
include_once( "strings.php" );
include_once( "jstream.php" );
include_once(     "log.php" );
include_once(     "ssl.php" );
include_once(   "mysql.php" );
include_once(   "input.php" );

$args = isset( $argv ) ? $argv : array();

main( $args );

function main( $argv )
{
    $start = microtime( TRUE );
    $delta = 0;

    NeoLog( "Starting new request:" . $_SERVER["REDIRECT_URL"] . "?" . urldecode( file_get_contents( 'php://input') ) );
    NeoLogIn();

    $environment = DetermineEnvironment         ( $argv                                                                         );
    $sp_name     = TranslateSPName              ( $environment["redirect_url"]                                                  );
                   CheckIfTunnelRequest         ( $sp_name,     $environment                                                    );
    $db_domain   = DetermineDatabaseLookupDomain( $environment["server_name"]                                                   );
    $db_info     = LookupDatabaseInfo           ( $db_domain                                                                    );
    $db_version  = LookupDatabaseVersion        ( $db_info                                                                      );
                   CheckIfDownloadRequest       ( $db_info,     $db_version, $sp_name, $environment                             );
    $sp_call     = MapRequestToSPCall           ( $db_info,     $db_version, $sp_name, $environment["request"]                  );
    $ret         = CallStoredProcedure          ( $db_info,     $db_version, $sp_name, $sp_call, $environment                   );
                   SetSessionIDCookie           ( $sp_name,     $ret                                                            );
    $response    = CreateResponse               ( $environment, $ret, ($delta = ceil( (microtime( TRUE ) - $start) * 1000000 )) );
                   OutputResponse               ( $sp_name,     $response, $environment["request"]                              );
                   LogToPubSub                  ( $environment, $db_info, $db_version, $sp_call, $response, $delta              );
}

function DetermineEnvironment( $argv )
{
    NeoLog( "Determining environment" );
    NeoLogIn();

    $environment = null;

    if ( "cli" != php_sapi_name() )
    {
        NeoLog( "Determining envionrment from Apache" );
        $environment = DetermineEnvironmentFromApache();
    }
    else
    {
        NeoLog( "Determining envionrment from command-line arguments" );
        $environment = DetermineEnvironmentFromArguments( $argv );
    }

    if ( ("" == $environment["server_name"]) || ("" == $environment["redirect_url"]) )
    {
        NeoLog( "Invalid environment, no SERVER_NAME or REDIRECT_URL." );
        NeoLog( "Commandline usage:" );
        NeoLog( "php index.php --server-name api.example.com --redirect-url /api/users/example/ --request '{\"param1\":\"value\",...}'" );
        exit;
    }

    if ( !defined( "DB_USERNAME" ) ) define( "DB_USERNAME", "public" );
    if ( !defined( "DB_PASSWORD" ) ) define( "DB_PASSWORD", "public" );
    if ( !defined( "SSL_BASE"    ) ) define( "SSL_BASE", "/etc/mysql/ssl/" );

    NeoLog( "Configuration: username: " . DB_USERNAME . "; password: " . DB_PASSWORD . "; ssl_base: " . SSL_BASE );

    NeoLogOut();

    return $environment;
}

function DetermineEnvironmentFromApache()
{
    $server_name  = "";
    $redirect_url = "";
    $requested    = "";

    if ( isset( $_SERVER ) && array_key_exists( "SERVER_NAME", $_SERVER ) )
    {
        $server_name = $_SERVER["SERVER_NAME"];
    }

    if ( isset( $_SERVER ) && array_key_exists( "REDIRECT_URL", $_SERVER ) )
    {
        $redirect_url = CanonicalisePath( $_SERVER["REDIRECT_URL"] );
    }

    if ( isset( $_REQUEST ) )
    {
        $request = string_has_prefix( $redirect_url, "/raw/" ) ? $_REQUEST : Input::FilterInput( $_REQUEST, null );
    }

    return array
    (
        "server_name"  => $server_name,
        "redirect_url" => $redirect_url,
        "request"      => $request
    );
}

function CanonicalisePath( $constant )
{
    if ( '.php' == substr( $constant, -4 ) )
    {
        $constant = dirname( $constant );
    }

    if ( '/' != substr( $constant, -1 ) )
    {
        $constant .= "/";
    }

    return $constant;
}

function DetermineEnvironmentFromArguments( $argv )
{
    $server_name  = "";
    $redirect_url = "";
    $request      = "";

    $next_is_server_name  = 0;
    $next_is_redirect_url = 0;
    $next_is_request      = 0;
    foreach( $argv as $arg )
    {
        if ( $next_is_server_name )
        {
            $server_name         = $arg;
            $next_is_server_name = 0;
        }
        else
        if ( $next_is_redirect_url )
        {
            $redirect_url         = $arg;
            $next_is_redirect_url = 0;
        }
        else
        if ( $next_is_request )
        {
            $request         = json_decode( $arg, TRUE );
            $next_is_request = 0;
        }
        else
        {
            switch ( $arg )
            {
            case "--server-name":
                $next_is_server_name = 1;
                break;

            case "--redirect-url":
                $next_is_redirect_url = 1;
                break;

            case "--request":
                $next_is_request = 1;
                break;
            }
        }
    }

    if ( !$request ) $request = array();

    return array
    (
        "server_name"  => $server_name,
        "redirect_url" => $redirect_url,
        "request"      => $request
    );
}

function TranslateSPName( $redirect_url )
{
    NeoLog( "Translating stored procedure name from url" );

    $bits = explode( '/', $redirect_url );

    if ( "" == $bits[0] ) $bits = array_slice( $bits, 1 );

    switch ( $bits[0] )
    {
    case 'api':
        $sp_name = trim( implode( '_', array_slice( $bits, 1 ) ), '_' );
        break;

    case 'auth':
        $sp_name = trim( implode( '_', array_slice( $bits, 0 ) ), '_' );
        break;

    case 'download':
        $sp_name = 'download';
        break;

    case 'raw':
        $sp_name = trim( implode( '_', array_slice( $bits, 0 ) ), '_' );
        break;

    case 'tunnel':
        $sp_name = 'tunnel';
        break;

    default:
        header( "Content-Type: text/plain" );
        echo "This is a Web API Bridge HTTP endpoint -- see https://www.webapibridge.org";
        exit;
    }

    return $sp_name;
}

function CheckIfTunnelRequest( $sp_name, $environment )
{
    if ( "tunnel" != $sp_name )
    {
        //
        //  This is mysteriously not being executed...
        //

        NeoLog( "Checking if tunnel request - NO" );
    }
    else
    {
        NeoLog( "Checking if tunnel request - YES" );
        NeoLogIn();
        Tunnel( $environment );
        NeoLogOut();
    }
}

function Tunnel( $environment )
{
    $request       = $environment['request'];
    $method        = array_get( $request, "wab_method"        );
    $authorization = array_get( $request, "wab_authorization" );
    $host          = array_get( $request, "wab_host"          );
    $endpoint      = array_get( $request, "wab_endpoint"      );
    $accept        = array_get( $request, "wab_accept"        );
    $content_type  = array_get( $request, "wab_content_type"  );
    $content64     = array_get( $request, "wab_content64"     );

    NeoLog( "WAB tunnel debug: $host$endpoint" );

    if ( ! ($method && $host && $endpoint && $accept && $content_type) )
    {
        header( "Content-Type: text/plain" );

        echo "Error: incomplete headers"         . "\n";
        echo "+ wab_method:       $method"       . "\n";
        echo "+ wab_host:         $host"         . "\n";
        echo "+ wab_endpoint:     $endpoint"     . "\n";
        echo "+ wab_accept:       $accept"       . "\n";
        echo "+ wab_content_type: $content_type" . "\n";
        echo "+ wab_content64:    $content64"    . "\n";
    }
    else
    {
        $curl         = SetupCURL( $method, $authorization, $host, $endpoint, $content_type, $content64, $accept );
        $response     = curl_exec ( $curl );
        $err          = curl_error( $curl );

        if ( $err )
        {
            header( "Content-Type: text/plain" );
            echo $err;
        }
        else
        {
            header( "Content-Type: $accept" );
            echo $response;
        }
    }

    ob_flush();
    flush();
    exit;
}

function SetupCURL( $method, $authorization, $host, $endpoint, $content_type, $content64, $accept )
{
    $curl = curl_init();

    $headers = array
    (
        "Accept: $accept",
        "Content-Type: $content_type",
        "Cache-Control: no-cache"
    );

    if ( "" != $authorization )
    {
        $headers[] = "Authorization: $authorization";
    }

    $url      = $host . $endpoint;
    $content  = base64_decode( $content64 );

    if ( "GET" == $method )
    {
        $url     = $url . "?" . $content;
        $content = "";
    }

    error_log( "WAB Tunnel: $url" );

    curl_setopt_array(
        $curl,
        array
        (
            CURLOPT_URL            => $url,
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING       => "",
            CURLOPT_MAXREDIRS      => 10,
            CURLOPT_TIMEOUT        => 30,
            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST  => $method,
            CURLOPT_POSTFIELDS     => $content,
            CURLOPT_HTTPHEADER     => $headers,
        )
    );

    return $curl;
}

function DetermineDatabaseLookupDomain( $server_name )
{
    if ( string_has_suffix( $server_name, ".local" ) )
    {
        $db_domain = "local-db." . str_replace( '.local', '', $server_name );
    }
    else
    {
        $db_domain = "db." . $server_name;
    }

    NeoLog( "Determining database lookup domain name: $db_domain" ); 

    return $db_domain;
}

function LookupDatabaseInfo( $db_domain )
{
    $db_info = null;
    $db_temp = (object) array( "dbname" => null, "hosts" => null );

    if ( function_exists( "apcu_exists" ) && apcu_exists( $db_domain ) )
    {
        $db_cached = apcu_fetch( $db_domain );
        $db_name   = $db_cached->dbname;
        $db_host   = $db_cached->hosts[0];

        if ( $db_name && $db_host )
        {
            NeoLog( "Looking up database info: found cached: $db_name@$db_host" );
            $db_info = $db_cached;
        }
    }

    if ( ! $db_info )
    {
        try
        {
            $records = dns_get_record( $db_domain, DNS_TXT );
            $n       = (null != $records) ? count( $records ) : 0;

            if ( 0 < $n )
            {
                $db_dns = ExtractDBInfoFromDNSRecords( $records );
            }
            else
            {
                $db_dns = ExtractDBInfoFromDBDomain( $db_domain );
            }

            $db_name = $db_dns->dbname;
            $db_host = $db_dns->hosts[0];

            if ( $db_name && $db_host )
            {
                NeoLog( "Looking up database info: DNS returned $n records: $db_name@$db_host" );
                $db_info = $db_dns;
            }

            if ( $db_info && function_exists( "apcu_add" ) )
            {
                apcu_add( $db_domain, $db_info );
                NeoLogIn();
                NeoLog( "Caching: added to cache: $db_domain" );
                NeoLogOut();
            }
        }
        catch ( Exception $ex )
        {
            NeoLog( "Exception while retrieving from DNS" );
            exit;
        }
    }

    NeoLogIn();
    if ( ! $db_info )
    {
        NeoLog( "Error, Could not determine database information, exiting." );
        exit;
    }
    else
    {
        //  Following will abort if cannot retrieve database connection.
        CheckDatabaseConnection( $db_info );
    }
    NeoLogOut();

    return $db_info;
}

function ExtractDBInfoFromDNSRecords( $records )
{
    $dbname = null;
    $hosts  = null;

    foreach ( $records as $record )
    {
        $host    = array_key_exists(    "host", $record ) ? $record[   "host"] : "";
        $class   = array_key_exists(   "class", $record ) ? $record[  "class"] : "";
        $ttl     = array_key_exists(     "ttl", $record ) ? $record[    "ttl"] : "";
        $type    = array_key_exists(    "type", $record ) ? $record[   "type"] : "";
        $txt     = array_key_exists(     "txt", $record ) ? $record[    "txt"] : "";

        $bits = explode( "@", $txt );
        
        switch ( count( $bits ) )
        {
        case 2:
            $dbname = $bits[0];
            $hosts  = explode( ",", $bits[1] );
            break;
        
        case 1:
            $hosts  = explode( ",", $bits[0] );
            break;
        }
    }

    return (object) array( "dbname" => $dbname, "hosts" => $hosts );
}

function ExtractDBInfoFromDBDomain( $db_domain )
{
    $truncated = str_replace( "www.", "", $db_domain );
    $truncated = str_replace( "api-", "", $truncated );
    $truncated = str_replace( "api.", "", $truncated );
    $bits      = explode( ".", $truncated );

    $n = count( $bits );

    for ( $i = 1; $i < $n; $i++ )
    {
        if ( ! IsTLD( $bits[$i], $i ) )
        {
            $dbname = $bits[$i];
            break;
        }
    }

    return (object) array( "dbname" => $dbname, "hosts" => array("localhost") );
}

function IsTLD( $domain_part, $level = 1 )
{
	switch ( $level )
	{
	case "1":
		switch ( $domain_part )
		{
		case "au":
		case "com":
		case "net";
		case "org";
		case "co":
		case "info":
		case "xyz":
		case "local":
			return true;
			
		default:
			return false;
		}
		break;

	case "2":
		switch ( $domain_part )
		{
		case "com":
		case "net";
		case "org";
		case "id":
		case "conf":
			return true;
			
		default:
			return false;
		}
		break;
	
	default:
		return false;
	}
}

function CheckDatabaseConnection( $db_info )
{
    if ( !defined( "DB_HOSTNAME" ) ) define( "DB_HOSTNAME", $db_info->hosts[0] );

    NeoLog( "Checking database connection to: " . $db_info->hosts[0] );
    NeoLogIn();

    $m = MySQLCreate();

    NeoLogOut();

    return TRUE;
}

function LookupDatabaseVersion( $db_info )
{
	$db_version = "";
    $db_name    = "";
    $db_hosts   = implode( ', ', $db_info->hosts );
    $db_host    = $db_info->hosts[0];

    $mysql  = MySQLCreate();
    $tuples = MySQLInfo( $mysql, "SHOW DATABASES" );

    foreach( $tuples as $tuple )
    {
        $tmp = $tuple["Database"];

        if ( string_has_prefix( $tmp, $db_info->dbname ) )
        {
            $db_name = $tmp;
        }
    }
    $db_version = $db_name ? str_replace( $db_info->dbname, "", $db_name ) : $db_version;

    if ( $db_name )
    {
        NeoLog( "Determining target database version: found: $db_name" );
    }
    else
    {
        NeoLog( "Determining target database version: cound not find database for: " . $db_info->dbname );
        exit;
    }

    return $db_version;
}

function CheckIfDownloadRequest( $db_info, $db_version, $sp_name, $environment )
{
    NeoLog( "Checking if tunnel request" );

    if ( "download" != $sp_name )
    {
        //
        //  This is mysteriously not being executed...
        //

        NeoLog( "Checking if download request - NO" );
    }
    else
    {
        NeoLog( "Performing download" );

        NeoLogIn();
        Download( $db_info, $db_version, $environment );
        NeoLogOut();
    }
}

function Download( $db_info, $db_version, $environment )
{
    $request = $environment['request'];
    $db      = $db_info->dbname . $db_version;
    $sid     = array_get( $request, "sid"   );
    $token   = array_get( $request, "token" );

    if ( $token )
    {
        $file = Retrieve( $db, $sid, $token );

        if ( $file )
        {
            $name    = $file->filename;
            $type    = $file->filetype;
            $size    = $file->filesize;
            $base64  = $file->base64;
            $content = base64_decode( $base64 );

            header( "Content-Type: application/octet-stream"                    );
            header( "Content-Length: $size"                                     );
            header( "Content-Disposition: attachment; filename=\"$name\""       );
            header( "Pragma: public"                                            );
            header( "Cache-Control: must-revalidate, post-check=0, pre-check=0" );
            echo $content;
        }
        else
        {
            header( "Content-Type: text/plain" );
            echo "File not found";
        }
    }
    else
    {
        header( "Content-Type: text/plain" );
        echo "No file specified";
    }

    ob_flush();
    flush();
    exit;
}

function Retrieve( $db, $sid, $token )
{
    $procedure = "Base_Files_Retrieve_By_Token( '$sid', '$token' )";

    NeoLog( "Retrieve: CALL $procedure" );

    $ret = MySQLProcedure( $db, $procedure );

    return array_key_exists( 0, $ret ) ? $ret[0] : null;
}

function MapRequestToSPCall( $db_info, $db_version, $sp_name, $request )
{
    $sp_call = null;

    if ( "multiselect" != $sp_name )
    {
        NeoLog( "Determine Stored Procedure Call" );
        NeoLogIn();


        $db = $db_info->dbname . $db_version;

        //
        //  Create list of provisional sp names.
        //

        $proc_name = null;
        $provisional[] = $sp_name;
        if ( string_has_suffix( $sp_name, "_csv" ) )
        {
            $provisional[] = preg_replace( "/_csv\z/", "", $sp_name );
        }
        $provisional[] = $provisional[count($provisional)-1] . "_retrieve";

        //
        //  Use first procedure found that matches provisional name.
        //

        $result = null;
        foreach( $provisional as $prov )
        {
            $sql = "RetrieveParametersFor( '$db', '$prov' )";
            
            $result = MySQLFunction( $db, $sql );

            if ( null !== $result )
            {
                $proc_name = $prov;
                break;
            }
        }

        if ( !$proc_name )
        {
            NeoLog( "Could not find procedure for: " . $sp_name );
            exit;
        }
        else
        {
            $arguments = "( ";

            $parameters = ConvertParametersToDictionary( $result );

            foreach ( $parameters as $parameter )
            {
                // Strip of leading '$' and make lowercase.
                $parameter_name = str_replace( '$', '', $parameter );
                $value          = array_get( $request, $parameter_name );

                if ( ! $value ) $value = array_get( $request, strtolower( $parameter_name ) );

                $arguments     .= ("'" . $value . "', ");

                if ( defined( "VERBOSE" ) && VERBOSE )
                {
                    $len = strlen( $value );
                    error_log( "$parameter_name ($parameter) --> ($len) $value" );
                }
            }
            
            $arguments = substr( $arguments, 0, -2 );
            
            $arguments .= " )";

            $sp_call = $proc_name . $arguments;

            if ( defined( "LOG_CALL" ) && LOG_CALL )
            {
                NeoLog( "SP Call: " . $sp_call );
            }
        }

        NeoLogOut();
    }

    return $sp_call;
}

function ConvertParametersToDictionary( $string )
{
    $parameters = array();
    $bits       = explode( ",", trim( $string ) );
    
    foreach ( $bits as $parameter )
    {
        $parts = explode( " ", trim( $parameter ) );
        
        $parameters[] = $parts[0];

        if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $parts[0] . " " . $parts[1] );
    }
    
    return $parameters;
}

function CallStoredProcedure( $db_info, $db_version, $sp_name, $sp_call, $environment )
{
    $ret = null;

    NeoLog( "Call Stored Procedure" );
    NeoLogIn();

    $db = $db_info->dbname . $db_version;

    if ( "multiselect" == $sp_name )
    {
        $ret = Multiselect   ( $db, $environment );
    }
    else
    {
        $ret = MySQLProcedure( $db, $sp_call );
    }

    NeoLogOut();

    return $ret;
}

function Multiselect( $db, $environment )
{
    $set        = array();
    $kinds      = explode( ',', array_get( $environment['request'], "kinds" ) );
    $sid        = array_get( $environment['request'], "sid"    );
    $filter     = array_get( $environment['request'], "filter" );

    foreach ( $kinds as $kind )
    {
        $bits = explode( ":", $kind );
        switch ( count( $bits ) )
        {
        case 1:
            $id          = "";
            $select_name = $bits[0];
            break;

        case 2:
            $id          = explode( ':', $kind )[0];
            $select_name = explode( ':', $kind )[1];
            break;

        default:
            continue;
        }

        $select_call = "selects_$select_name( '$sid', '$id', '', '$filter' )";

        $tuples      = MySQLProcedure( $db, $select_call );

        if ( is_array( $tuples ) )
        {
            $set[] = (object) array( "id" => $id, "name" => $kind, "tuples" => $tuples, "call" => $select_call );
        }
        else
        {
            $set[] = (object) array( "id" => $id, "name" => $kind,  "error" => $tuples, "call" => $select_call );
        }
    }

    return $set;
}

function SetSessionIDCookie( $sp_name, $ret )
{
    NeoLog( "Existing sessionid: " . array_get( $_COOKIE, "sid" ) );

    if ( "auth_login" == $sp_name )
    {
        if ( 1 == count( $ret ) )
        {
            $tuple     = $ret[0];
            $sessionid = $tuple->sessionid;
            $cookie    = "Set-Cookie: sid=";
            $cookie    = $cookie . $sessionid;
            $cookie    = $cookie . "; path=/; HttpOnly";
        
            header( $cookie );

            NeoLog( "Setting Session ID Cookie: " . $sessionid );
        }
    }
}

function CreateResponse( $environment, $ret, $microseconds )
{
    $request = $environment["request"];

    $response              = array();
    $response['URL'      ] = $environment["redirect_url"];
    $response['error'    ] = !is_array( $ret ) ? $ret : "";
    $response['failover' ] = "FALSE";
    $response['hostname' ] = "";
    $response['message'  ] = "";
    $response['status'   ] = !is_array( $ret ) ? "ERROR" : "OK";
    $response['warning'  ] = "";
    $response['target_id'] = array_get( $request, "target_id" );
    $response['submit'   ] = array_get( $request, "submit"    );
    $response['limit'    ] = array_get( $request, "limit"     );
    $response['offset'   ] = array_get( $request, "offset"    );
    $response['microsecs'] = $microseconds;
    $response['results'  ] = is_array( $ret ) ? $ret : NULL;

    return (object) $response;
}

function OutputResponse( $sp_name, $response, $request )
{
    NeoLog( "Output Response" );
    NeoLogIn();

    if ( string_has_suffix( $sp_name, "_csv" ) )
    {
        TranslateToCSV ( $response, $request );
    }
    else
    {
        TranslateToJSON( $response );
    }

    NeoLogOut();
}

function TranslateToJSON( $results )
{
    NeoLog( "Translate to JSON" );

    if ( ob_get_contents() ) ob_end_clean();

    set_time_limit(0);
    ignore_user_abort( TRUE );

    ob_start();
    {
        \JStream::Encode( 0, $results );
        echo "\n";
        $size = ob_get_length();

        header( "Content-Length: $size"          );
        header( "Connection: close"              );
        header( "Content-Encoding: none"         );
        header( "Content-Type: application/json" );
    }
    ob_end_flush();
    ob_flush();
    flush();

    if (session_id()) session_write_close();
}

function TranslateToCSV( $response, $request )
{
    NeoLog( "Translate to CSV" );

    if ( "OK" == $response->status )
    {
        $content  = "";
        $ds       = ("true" == array_get( $request, "double_space" ));
        $numbered = ("true" == array_get( $request, "numbered"     ));
        $map      = GenerateMap( array_get( $request, "csv_map" ) );

        if ( $map )
        {
            NeoLog( "Calling CSV::encode with mapped tuples." );

            $content = \CSV::encode( \CSV::MapValues( $map, $response->results ), $numbered, $ds );
        }
        else
        {
            NeoLog( "Calling CSV::encode with unmapped tuples." );

            $content = \CSV::encode( $response->results, $numbered, $ds );
        }

        $size = strlen( $content );

        if ( false )
        {
            error_log( "Content-Type: application/octet-stream" );
            error_log( "Content-Length: $size" );
            error_log( "Content-Disposition: attachment;filename=\"download\"" );
        }

        header( "Content-Type: application/octet-stream"                    );
        header( "Content-Length: $size"                                     );
        header( "Content-Disposition: attachment;filename=\"download.csv\"" );
        header( "Pragma: public"                                            );
        header( "Cache-Control: must-revalidate, post-check=0, pre-check=0" );
        header( "Connection: close" );
        echo $content;

        flush();
        exit;
    }
}

function GenerateMap( $base64_encoded_csv_map )
{
    $csv_map =  base64_decode( $base64_encoded_csv_map );

    if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "GenerateMap( \"$csv_map\" )" );

    //
    //  csv_map=Family+Name|family_name,...
    //

    $map = null;

    if ( $csv_map )
    {
        $map  = array();
        $bits = explode( ",", $csv_map );
        
        foreach ( $bits as $mapping )
        {
            $pieces = explode( "|", $mapping );

            if ( 2 == count( $pieces ) )
            {
                $key = urldecode( $pieces[0] );
                $val = urldecode( $pieces[1] );
            
                $map[$key] = $val;
            }

            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "Mapping( $mapping ) : $key | $val" );
        }
    }
    return $map;
}

use Google\Cloud\PubSub\PubSubClient;

function LogToPubSub( $environment, $db_info, $db_version, $sp_call, $response, $delta )
{
    if ( "TRUE" == getenv( "USE_GOOGLE_PUB_SUB" ) )
    {
        $topic = GetPubSubTopic();

        if ( !$topic )
        {
            NeoLog( "Not logged to Google PubSub" );
        }
        else
        {
            $now          = date( "c", time() );
            $api_hostname = $environment['server_name' ];
            $endpoint     = $environment['redirect_url'];
            $request      = $environment['request'     ];
            $db           = $db_info->dbname . $db_version;
            $db_hostname  = DB_HOSTNAME;
            $sql64        = base64_encode( $sp_call );
            $ip_address   = $_SERVER['SERVER_ADDR'];

            $requesting_url           = array_get(  $request, "wab_requesting_url" );

            $cgi_request_method       = array_get( $_SERVER , "REQUEST_METHOD"       );
            $cgi_request_time         = array_get( $_SERVER , "REQUEST_TIME"         );
            $cgi_http_accept          = array_get( $_SERVER , "HTTP_ACCEPT"          );
            $cgi_http_accept_charset  = array_get( $_SERVER , "HTTP_ACCEPT_CHARSET"  );
            $cgi_http_accept_encoding = array_get( $_SERVER , "HTTP_ACCEPT_ENCODING" );
            $cgi_http_accept_language = array_get( $_SERVER , "HTTP_ACCEPT_LANGUAGE" );
            $cgi_http_referer         = array_get( $_SERVER , "HTTP_REFERER"         );
            $cgi_http_user_agent      = array_get( $_SERVER , "HTTP_USER_AGENT"      );
            $cgi_remote_addr          = array_get( $_SERVER , "HTTP_REMOTE_ADDR"     );
            $cgi_remote_port          = array_get( $_SERVER , "HTTP_REMOTE_PORT"     );
            $cgi_remote_user          = array_get( $_SERVER , "HTTP_REMOTE_USER"     );

            $json  = "{\n";
            $json .= "\"ts\":                       \"$now\",\n";
            $json .= "\"api_hostname\":             \"$api_hostname\",\n";
            $json .= "\"endpoint\":                 \"$endpoint\",\n";
            $json .= "\"type\":                     \"r/w\",\n";
            $json .= "\"status\":                   \"$response->status\",\n";
            $json .= "\"error\":                    \"$response->error\",\n";
            $json .= "\"sql64\":                    \"$sql64\",\n";
            $json .= "\"response_time_ms\":         \"$response->microsecs\",\n";
            $json .= "\"db\":                       \"$db\",\n";
            $json .= "\"db_version\":               \"$db_version\",\n";
            $json .= "\"db_hostname\":              \"$db_hostname\",\n";
            $json .= "\"ip_address\":               \"$ip_address\",\n";
            $json .= "\"requesting_url\":           \"$requesting_url\",\n";
            $json .= "\"cgi_request_method\":       \"$cgi_request_method\",\n";
            $json .= "\"cgi_request_time\":         \"$cgi_request_time\",\n";
            $json .= "\"cgi_http_accept\":          \"$cgi_http_accept\",\n";
            $json .= "\"cgi_http_accept_charset\":  \"$cgi_http_accept_charset\",\n";
            $json .= "\"cgi_http_accept_encoding\": \"$cgi_http_accept_encoding\",\n";
            $json .= "\"cgi_http_accept_language\": \"$cgi_http_accept_language\",\n";
            $json .= "\"cgi_http_referer\":         \"$cgi_http_referer\",\n";
            $json .= "\"cgi_http_user_agent\":      \"$cgi_http_user_agent\",\n";
            $json .= "\"cgi_remote_addr\":          \"$cgi_remote_addr\",\n";
            $json .= "\"cgi_remote_port\":          \"$cgi_remote_port\",\n";
            $json .= "\"cgi_remote_user\":          \"$cgi_remote_user\"\n";
            $json .= "}\n";

            try
            {
                $topic->publish( ['data' => $json] );

                NeoLog( "Logged to Google PubSub" );
            }
            catch ( Exception $ex )
            {
                NeoLog( $ex );
            }
        }
    }
}

function GetPubSubTopic()
{
    $topic = null;

    NeoLogIn();

    if ( !getenv( "GOOGLE_APPLICATION_CREDENTIALS" ) )
    {
        NeoLog( "!!! WARNING !!! GOOGLE_APPLICATION_CREDENTIALS not defined in Apache configuration, but USE_GOOGLE_PUB_SUB is!" );
    }
    else
    if ( !getenv( "GOOGLE_PUB_SUB_PROJECT" ) )
    {
        NeoLog( "!!! WARNING !!! GOOGLE_PUB_SUB_PROJECT not defined in Apache configuration, but USE_GOOGLE_PUB_SUB is!" );
    }
    else
    if ( !getenv( "GOOGLE_PUB_SUB_TOPIC" ) )
    {
        NeoLog( "!!! WARNING !!! GOOGLE_PUB_SUB_TOPIC not defined in Apache configuration, but USE_GOOGLE_PUB_SUB is!" );
    }
    else
    {
        $pubsub_project = getenv( "GOOGLE_PUB_SUB_PROJECT" );
        $pubsub_topic   = getenv( "GOOGLE_PUB_SUB_TOPIC"   );

        try
        {
            $pubsub = new PubSubClient( ['projectId' => $pubsub_project] );
            $topic  = $pubsub->topic( $pubsub_topic );
        }
        catch( Exception $e )
        {
            NeoLog( "!!! WARNING !!! Could not use GOOGLE credentials: " . getEnv( "GOOGLE_APPLICATION_CREDENTIALS" ) );
        }
    }

    NeoLogOut();

    return $topic;
}

